import torch
from .base_model import BaseModel
from . import networks
import torchvision
import torch.nn as nn
from models.vit import ViT
from models.fcn_resnet50 import FCNResnet50
import torch.nn.functional as F
from models.ait import SwinV2TransformerRPE2FC
import torchvision.transforms as transforms
import kornia

class Pix2PixModel(BaseModel):
    """ This class implements the pix2pix model, for learning a mapping from input images to output images given paired data.

    The model training requires '--dataset_mode aligned' dataset.
    By default, it uses a '--netG unet256' U-Net generator,
    a '--netD basic' discriminator (PatchGAN),
    and a '--gan_mode' vanilla GAN loss (the cross-entropy objective used in the orignal GAN paper).

    pix2pix paper: https://arxiv.org/pdf/1611.07004.pdf
    """
    @staticmethod
    def modify_commandline_options(parser, is_train=True):
        """Add new dataset-specific options, and rewrite default values for existing options.

        Parameters:
            parser          -- original option parser
            is_train (bool) -- whether training phase or test phase. You can use this flag to add training-specific or test-specific options.

        Returns:
            the modified parser.

        For pix2pix, we do not use image buffer
        The training objective is: GAN Loss + lambda_L1 * ||G(A)-B||_1
        By default, we use vanilla GAN loss, UNet with batchnorm, and aligned datasets.
        """
        # changing the default values to match the pix2pix paper (https://phillipi.github.io/pix2pix/)
        parser.set_defaults(norm='batch', netG='unet_256', dataset_mode='aligned')
        if is_train:
            parser.set_defaults(pool_size=0, gan_mode='vanilla')
            parser.add_argument('--lambda_L1', type=float, default=100.0, help='weight for L1 loss')

        return parser

    def __init__(self, opt):
        """Initialize the pix2pix class.

        Parameters:
            opt (Option class)-- stores all the experiment flags; needs to be a subclass of BaseOptions
        """
        BaseModel.__init__(self, opt)
        # specify the training losses you want to print out. The training/test scripts will call <BaseModel.get_current_losses>
        self.loss_names = ['Distill', 'G_GAN', 'G_L1', 'D_real', 'D_fake']
        # specify the images you want to save/display. The training/test scripts will call <BaseModel.get_current_visuals>
        self.visual_names = ['real_A', 'fake_B', 'real_B']
        # specify the models you want to save to the disk. The training/test scripts will call <BaseModel.save_networks> and <BaseModel.load_networks>
        if self.isTrain:
            self.model_names = ['G', 'D']
        else:  # during test time, only load G
            self.model_names = ['G']
        # define networks (both generator and discriminator)
        self.netG = networks.define_G(opt.input_nc, opt.output_nc, opt.ngf, opt.netG, opt.norm,
                                      not opt.no_dropout, opt.init_type, opt.init_gain, self.gpu_ids)

        # self.teacher_task = opt.teacher_task
        self.teacher_task = "insseg"
        assert self.teacher_task in ['random', 'classif', 'depth', 'insseg']

        # self.projection_direction = opt.projection_direction
        self.projection_direction = "t2s"
        assert self.projection_direction in ["s2t", "t2s"]

        self.model  = opt.model 

        if self.teacher_task ==  "classif":
            self.center_crop = torchvision.transforms.CenterCrop((224, 224))
            self.teacher = ViT()

            if self.projection_direction == "s2t":
                self.projection = nn.Linear(1024, 768) 
            elif self.projection_direction ==  "t2s":
                self.projection = nn.Linear(768, 1024) 
            else:
                raise NotImplementedError()

        elif self.teacher_task in ["random", "insseg"]:

            if self.teacher_task ==  "random":
                self.teacher = FCNResnet50(random_weights=True)
            else:
                self.teacher = FCNResnet50(random_weights=False)

            if self.projection_direction == "s2t":
                self.projection = nn.Conv2d(512, 2048, kernel_size=1)
            elif self.projection_direction ==  "t2s":
                self.projection = nn.Conv2d(2048, 512, kernel_size=1) 
            else:
                raise NotImplementedError()

        elif self.teacher_task == "depth":
            model_kwargs = {
                "pretrain_img_size":192,
                "embed_dim":128,
                "depths":[2, 2, 18, 2],
                "num_heads":[4, 8, 16, 32],
                "window_size":[30, 30, 30, 15],
                "use_shift":[True, True, False, False],
                "pretrain_window_size":[12, 12, 12, 6],
                "ape":False,
                "drop_path_rate":0.1,
                "patch_norm":True,
                "use_checkpoint":False,
                "out_indices":(3,)
            }
            ckpt_path = "./ckpts/ait_depth_swinv2b_ar.pth"
            weights = torch.load(ckpt_path)['state_dict']

            weights = {key: val for key, val in weights.items() if key.startswith("backbone.")}
            weights = {key.replace("backbone.", ""): val for key, val in weights.items()}

            self.teacher = SwinV2TransformerRPE2FC(**model_kwargs)
            self.teacher.load_state_dict(weights)
            
            if self.projection_direction == "s2t":
                self.projection = nn.Conv2d(1024, 1024, kernel_size=1)
            elif self.projection_direction ==  "t2s":
                self.projection = nn.Conv2d(1024, 1024, kernel_size=1) 
            else:
                raise NotImplementedError()     

        self.teacher.cuda()
        self.teacher.eval()
        self.projection.cuda()
        self.loss_Distill = 0.0

        if self.isTrain:  # define a discriminator; conditional GANs need to take both input and output images; Therefore, #channels for D is input_nc + output_nc
            self.netD = networks.define_D(opt.input_nc + opt.output_nc, opt.ndf, opt.netD,
                                          opt.n_layers_D, opt.norm, opt.init_type, opt.init_gain, self.gpu_ids)

        if self.isTrain:
            # define loss functions
            self.criterionGAN = networks.GANLoss(opt.gan_mode).to(self.device)
            self.criterionL1 = torch.nn.L1Loss()
            # initialize optimizers; schedulers will be automatically created by function <BaseModel.setup>.
            trainable_list = list(self.netG.parameters()) + list(self.projection.parameters())
            self.optimizer_G = torch.optim.Adam(trainable_list, lr=opt.lr, betas=(opt.beta1, 0.999))
            self.optimizer_D = torch.optim.Adam(self.netD.parameters(), lr=opt.lr, betas=(opt.beta1, 0.999))
            self.optimizers.append(self.optimizer_G)
            self.optimizers.append(self.optimizer_D)

    def set_input(self, input):
        """Unpack input data from the dataloader and perform necessary pre-processing steps.

        Parameters:
            input (dict): include the data itself and its metadata information.

        The option 'direction' can be used to swap images in domain A and domain B.
        """
        AtoB = self.opt.direction == 'AtoB'
        self.real_A = input['A' if AtoB else 'B'].to(self.device)
        self.real_B = input['B' if AtoB else 'A'].to(self.device)
        self.image_paths = input['A_paths' if AtoB else 'B_paths']

    def lab2rgb(self, L, AB):
        """Convert an Lab tensor image to a RGB numpy output
        Parameters:
            L  (1-channel tensor array): L channel images (range: [-1, 1], torch tensor array)
            AB (2-channel tensor array):  ab channel images (range: [-1, 1], torch tensor array)

        Returns:
            rgb (RGB numpy image): rgb output images  (range: [0, 255], numpy array)
        """
        AB2 = AB * 110.0
        L2 = (L + 1.0) * 50.0
        Lab = torch.cat([L2, AB2], dim=1)
        Lab = Lab[0].data.cpu().float().numpy()
        Lab = np.transpose(Lab.astype(np.float64), (1, 2, 0))
        rgb = color.lab2rgb(Lab) * 255
        return rgb

    def lab2rgb_v2(self, L, AB):
        AB2 = AB * 110.0
        L2 = (L + 1.0) * 50.0
        Lab = torch.cat([L2, AB2], dim=1)
        rgb = kornia.color.lab_to_rgb(Lab)
        return rgb

    def forward(self):
        """Run forward pass; called by both functions <optimize_parameters> and <test>."""
        self.loss_Distill = 0.0
        self.fake_B, z_s = self.netG(self.real_A)  # G(A)

        if self.model == "colorization":
            normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                             std=[0.229, 0.224, 0.225])

            # very inefficient
            # teacher_input = torch.tensor(self.lab2rgb(self.real_A, self.real_B)).unsqueeze(0)
            # teacher_input = (teacher_input / 255.0)
            # teacher_input = torch.einsum('bhwc->bchw', teacher_input)

            # rgb input
            # teacher_input = self.lab2rgb_v2(self.real_A, self.real_B)
            # teacher_input = normalize(teacher_input).float().cuda()

            # grayscale input
            teacher_input = torch.tile(self.real_A, (1, 3, 1, 1))
            
        elif self.model == "pix2pix":
            # teacher_input = self.real_B
            teacher_input = self.real_A
        else:
            raise NotImplementedError()

        if self.teacher_task == "classif":
            # torch.Size([1, 1024, 16, 16])
            z_s = z_s[3]

            # vit: teacher forward pass
            # ([1, 768, 14, 14])
            # real_B_crop = self.center_crop(teacher_input)
            # z_t = self.teacher.forward_feats(real_B_crop)
            # z_t = self.teacher.extract_hw_feats(z_t)
            # z_t.detach()

            # b, c, h, w = z_s.shape
            z_s = z_s.view(b, c, h * w).mean(2)

            # b, c, h, w = z_t.shape
            # z_t = z_t.view(b, c, h * w).mean(2)

        elif self.teacher_task in ["random", "insseg"]:
            # torch.Size([1, 512, 32, 32])
            z_s = z_s[4]

            # z_t = self.teacher.forward_feats(self.real_A)
            z_t = self.teacher.forward_feats(teacher_input)
            z_t.detach()

        elif self.teacher_task == "depth":
            # torch.Size([1, 1024, 8, 8])
            z_s = z_s[2]

            # torch.Size([1, 1024, 8, 8])
            # z_t = self.teacher(self.real_A)
            z_t = self.teacher(teacher_input)
            z_t = z_t[-1]
            z_t.detach()
        else:
            raise NotImplementedError()

        if self.projection_direction ==  "s2t":
            z_s = self.projection(z_s)
        elif self.projection_direction == "t2s":
            z_t = self.projection(z_t)
        else:
            raise NotImplementedError()
            
        # different metrics
        # b, c, h, w = z_s.shape
        # self.loss_Distill = 200. * F.mse_loss(
        #     F.normalize(z_s.view(b, c, h * w), dim=2), 
        #     F.normalize(z_t.view(b, c, h * w), dim=2)
        # )
        # z_s_projected = (z_s_projected - z_s_projected.mean(0)) / z_s_projected.std(0)
        # z_t = (z_t - z_t.mean(0)) / z_t.std(0)

        self.loss_Distill = 2.0 * F.mse_loss(z_s, z_t)

    def backward_D(self):
        """Calculate GAN loss for the discriminator"""
        # Fake; stop backprop to the generator by detaching fake_B
        fake_AB = torch.cat((self.real_A, self.fake_B), 1)  # we use conditional GANs; we need to feed both input and output to the discriminator
        pred_fake = self.netD(fake_AB.detach())
        self.loss_D_fake = self.criterionGAN(pred_fake, False)
        # Real
        real_AB = torch.cat((self.real_A, self.real_B), 1)
        pred_real = self.netD(real_AB)
        self.loss_D_real = self.criterionGAN(pred_real, True)
        # combine loss and calculate gradients
        self.loss_D = (self.loss_D_fake + self.loss_D_real) * 0.5
        self.loss_D.backward()

    def backward_G(self):
        """Calculate GAN and L1 loss for the generator"""
        # First, G(A) should fake the discriminator
        fake_AB = torch.cat((self.real_A, self.fake_B), 1)
        pred_fake = self.netD(fake_AB)
        self.loss_G_GAN = self.criterionGAN(pred_fake, True)
        # Second, G(A) = B
        self.loss_G_L1 = self.criterionL1(self.fake_B, self.real_B) * self.opt.lambda_L1
        # combine loss and calculate gradients
        self.loss_G = self.loss_G_GAN + self.loss_G_L1
        # add distillation loss
        self.loss_G += self.loss_Distill
        self.loss_G.backward()

    def optimize_parameters(self):
        self.forward()                   # compute fake images: G(A)
        # update D
        self.set_requires_grad(self.netD, True)  # enable backprop for D
        self.optimizer_D.zero_grad()     # set D's gradients to zero
        self.backward_D()                # calculate gradients for D
        self.optimizer_D.step()          # update D's weights
        # update G
        self.set_requires_grad(self.netD, False)  # D requires no gradients when optimizing G
        self.optimizer_G.zero_grad()        # set G's gradients to zero
        self.backward_G()                   # calculate graidents for G
        self.optimizer_G.step()             # update G's weights
