"""General-purpose test script for image-to-image translation.

Once you have trained your model with train.py, you can use this script to test the model.
It will load a saved model from '--checkpoints_dir' and save the results to '--results_dir'.

It first creates model and dataset given the option. It will hard-code some parameters.
It then runs inference for '--num_test' images and save results to an HTML file.

Example (You need to train models first or download pre-trained models from our website):
    Test a CycleGAN model (both sides):
        python test.py --dataroot ./datasets/maps --name maps_cyclegan --model cycle_gan

    Test a CycleGAN model (one side only):
        python test.py --dataroot datasets/horse2zebra/testA --name horse2zebra_pretrained --model test --no_dropout

    The option '--model test' is used for generating CycleGAN results only for one side.
    This option will automatically set '--dataset_mode single', which only loads the images from one set.
    On the contrary, using '--model cycle_gan' requires loading and generating results in both directions,
    which is sometimes unnecessary. The results will be saved at ./results/.
    Use '--results_dir <directory_path_to_save_result>' to specify the results directory.

    Test a pix2pix model:
        python test.py --dataroot ./datasets/facades --name facades_pix2pix --model pix2pix --direction BtoA

See options/base_options.py and options/test_options.py for more test options.
See training and test tips at: https://github.com/junyanz/pytorch-CycleGAN-and-pix2pix/blob/master/docs/tips.md
See frequently asked questions at: https://github.com/junyanz/pytorch-CycleGAN-and-pix2pix/blob/master/docs/qa.md
"""
import os
from options.test_options import TestOptions
from data import create_dataset
from models import create_model
from util.visualizer import save_images
from util import html
from util.util import tensor2im
import torch.nn.functional as F
import torch
from torchmetrics import PeakSignalNoiseRatio
from torchmetrics.image.fid import FrechetInceptionDistance
from torchmetrics import StructuralSimilarityIndexMeasure
import torchvision
from torchmetrics.image.inception import InceptionScore 
_ = torch.manual_seed(123)

try:
    import wandb
except ImportError:
    print('Warning: wandb package cannot be found. The option "--use_wandb" will result in error.')


if __name__ == '__main__':
    opt = TestOptions().parse()  # get test options
    # hard-code some parameters for test
    opt.num_threads = 0   # test code only supports num_threads = 0
    opt.batch_size = 1    # test code only supports batch_size = 1
    opt.serial_batches = True  # disable data shuffling; comment this line if results on randomly chosen images are needed.
    opt.no_flip = True    # no flip; comment this line if results on flipped images are needed.
    opt.display_id = -1   # no visdom display; the test code saves the results to a HTML file.
    dataset = create_dataset(opt)  # create a dataset given opt.dataset_mode and other options
    model = create_model(opt)      # create a model given opt.model and other options
    model.setup(opt)               # regular setup: load and print networks; create schedulers

    # initialize logger
    if opt.use_wandb:
        wandb_run = wandb.init(project=opt.wandb_project_name, name=opt.name, config=opt) if not wandb.run else wandb.run
        wandb_run._label(repo='CycleGAN-and-pix2pix')

    # create a website
    web_dir = os.path.join(opt.results_dir, opt.name, '{}_{}'.format(opt.phase, opt.epoch))  # define the website directory
    if opt.load_iter > 0:  # load_iter is 0 by default
        web_dir = '{:s}_iter{:d}'.format(web_dir, opt.load_iter)
    print('creating web directory', web_dir)
    webpage = html.HTML(web_dir, 'Experiment = %s, Phase = %s, Epoch = %s' % (opt.name, opt.phase, opt.epoch))
    # test with eval mode. This only affects layers like batchnorm and dropout.
    # For [pix2pix]: we use batchnorm and dropout in the original pix2pix. You can experiment it with and without eval() mode.
    # For [CycleGAN]: It should not affect CycleGAN as CycleGAN uses instancenorm without dropout.
    if opt.eval:
        model.eval()

    psnr_fn = PeakSignalNoiseRatio().cuda()
    ssim_fn = StructuralSimilarityIndexMeasure().cuda()
    fid = FrechetInceptionDistance(feature=2048, normalize=True).cuda()
    inception = InceptionScore(normalize=True).cuda()
    center_crop = torchvision.transforms.CenterCrop((100, 100))
    total_loss = []
    total_psnr = []
    total_ssim = []
    img_real_array = []
    img_fake_array = []
    for i, data in enumerate(dataset):
        if i >= opt.num_test:  # only apply our model to opt.num_test images.
            break
        model.set_input(data)  # unpack data from data loader
        model.test()           # run inference

        img_real = data['B'].cuda()
        img_fake = model.fake_B

        psnr = psnr_fn(img_fake, img_real)
        total_psnr.append(psnr)

        cur_loss = F.mse_loss(F.normalize(img_fake, dim=1), F.normalize(img_real, dim=1))
        total_loss.append(cur_loss)

        ssim = ssim_fn(img_fake, img_real)
        total_ssim.append(ssim)

        # img_real = torch.tensor(tensor2im(data['B'].cuda())).unsqueeze(0).permute(0, 3, 1, 2).cuda()
        # img_fake = torch.tensor(tensor2im(model.fake_B)).unsqueeze(0).permute(0, 3, 1, 2).cuda()
        # img_real_array.append(img_real)
        # img_fake_array.append(img_fake)

        # img_real = center_crop(img_real)
        # img_fake = center_crop(img_fake)

        # inception.update(img_fake)
        # fid.update(img_real, real=True)
        # fid.update(img_fake, real=False)

        visuals = model.get_current_visuals()  # get image results
        img_path = model.get_image_paths()     # get image paths
        if i % 5 == 0:  # save images to an HTML file
            print('processing (%04d)-th image... %s' % (i, img_path))
        save_images(webpage, visuals, img_path, aspect_ratio=opt.aspect_ratio, width=opt.display_winsize, use_wandb=opt.use_wandb)

    # img_real = torch.cat(img_real_array, dim=0)
    # img_fake = torch.cat(img_fake_array, dim=0)
    # fid.update(img_real, real=True)
    # fid.update(img_fake, real=False)

    total_loss = torch.stack(total_loss).mean()
    total_psnr = torch.stack(total_psnr).mean()
    total_ssim = torch.stack(total_ssim).mean()
    # total_fid = fid.compute()
    # total_is = inception.compute()
    print(f"Total mse = {total_loss}")
    print(f"Total PSNR = {total_psnr}")
    # print(f"Total FID = {total_fid}")
    print(f"Total SSIM = {total_ssim}")
    # print(f"Total IS = {total_is}")
    webpage.save()  # save the HTML
