import os.path as osp

from mmcv.utils import TORCH_VERSION, digit_version
from mmcv.runner.checkpoint import save_checkpoint
from mmcv.runner.hooks import HOOKS
from mmcv.runner.utils import get_host_info
from mmcv.runner.dist_utils import master_only
from mmcv.runner.hooks.logger.base import LoggerHook
import matplotlib
import torch
import PIL.Image
import matplotlib.pyplot as plt

import csv
import numpy as np
import seaborn as sns
import pandas as pd

@HOOKS.register_module()
class CustomTensorboardLoggerHook(LoggerHook):
    """Class to log metrics to Tensorboard. Modified to work with matplotlib.figure.Figures.

    Args:
        log_dir (string): Save directory location. Default: None. If default
            values are used, directory location is ``runner.work_dir``/tf_logs.
        interval (int): Logging interval (every k iterations). Default: True.
        ignore_last (bool): Ignore the log of last iterations in each epoch
            if less than `interval`. Default: True.
        reset_flag (bool): Whether to clear the output buffer after logging.
            Default: False.
        by_epoch (bool): Whether EpochBasedRunner is used. Default: True.
    """

    def __init__(self,
                 log_dir=None,
                 interval=10,
                 svd_interval=100,
                 ignore_last=True,
                 reset_flag=False,
                 by_epoch=True):
        super(CustomTensorboardLoggerHook, self).__init__(interval, ignore_last,
                                                    reset_flag, by_epoch)
        self.svd_interval = svd_interval    # Used for SVD plots. Must divide by interval (or it won't be run)
        self.log_dir = log_dir
        self.svals = []

    @master_only
    def before_run(self, runner):
        super(CustomTensorboardLoggerHook, self).before_run(runner)
        if (TORCH_VERSION == 'parrots'
                or digit_version(TORCH_VERSION) < digit_version('1.1')):
            try:
                from tensorboardX import SummaryWriter
            except ImportError:
                raise ImportError('Please install tensorboardX to use '
                                  'CustomTensorboardLoggerHook.')
        else:
            try:
                from torch.utils.tensorboard import SummaryWriter
            except ImportError:
                raise ImportError(
                    'Please run "pip install future tensorboard" to install '
                    'the dependencies to use torch.utils.tensorboard '
                    '(applicable to PyTorch 1.1 or higher)')

        if self.log_dir is None:
            self.log_dir = osp.join(runner.work_dir, 'tf_logs')
        self.writer = SummaryWriter(self.log_dir)

    @master_only
    def log(self, runner):
        tags = self.get_loggable_tags(runner, allow_text=True)
        for tag, val in tags.items():
            if isinstance(val, str):
                self.writer.add_text(tag, val, self.get_iter(runner))
            elif isinstance(val, matplotlib.figure.Figure):
                self.writer.add_figure(tag, figure=val, global_step=self.get_iter(runner))
                runner.log_buffer.output.pop(tag)
            else:
                self.writer.add_scalar(tag, val, self.get_iter(runner))

        def gen_plot(svals_list_list):
            """Create a pyplot plot and save to buffer."""
            plt.close()
            fig, axs = plt.subplots(2, len(svals_list_list), squeeze=False, figsize=[6 * len(svals_list_list), 8])
            [ax.clear() for ax in axs.ravel()]
            for svals_idx, svals_list in enumerate(svals_list_list):
                sns.set_palette("coolwarm", n_colors=len(svals_list))
                axs[0, svals_idx].set_ylim([0, 1.1])
                for i, svals in enumerate(svals_list):
                    svals = svals.cpu().detach().numpy()
                    x = np.linspace(0, len(svals), len(svals))
                    data = pd.DataFrame({"Sorted singular value index": x, "Singular values of Wp": svals})
                    data['Normed Singular values of Wp'] = data['Singular values of Wp'] / data['Singular values of Wp'].max()

                    sns_plot = sns.lineplot(
                        x="Sorted singular value index", 
                        y="Normed Singular values of Wp", 
                        data=data,
                        ax=axs[0, svals_idx]
                    ).set(title=f"Evolvement of singular values of Wp (proj. {svals_idx})")

                    sns_plot = sns.lineplot(
                        x="Sorted singular value index", 
                        y="Singular values of Wp", 
                        data=data,
                        ax=axs[1, svals_idx]
                    ).set(title=f"Evolvement of singular values of Wp (unnormed) (proj. {svals_idx})")
                    # ylabels = ['{:,g}'.format(sval) for sval in axs[0, svals_idx].get_yticks()]
                    # axs[0, svals_idx].set_yticklabels(ylabels)

            fig.tight_layout()
            return fig

        if self.get_iter(runner) % self.svd_interval == 0:
            # add singular value plot
            if hasattr(runner.model.module, "projections"):
                if runner.model.module.projections is not None and len(runner.model.module.projections) > 0:
                    if len(self.svals) != len(runner.model.module.projections):
                        self.svals = [[] for _ in runner.model.module.projections]
                    assert runner.model.module.projection_arch in ["conv1x1"], "Error: projection arch not recognized"
                    # contract Kh=Kw=1 dim
                    for proj_idx in range(len(runner.model.module.projections)):
                        m = runner.model.module.projections[proj_idx].projection.weight[:, :, 0, 0]
                        svals = torch.linalg.svdvals(m)
                        approx_k =  torch.count_nonzero( torch.where(svals > 1e-6, svals, 0.0) )
                        # add svd vals for next log (so we can capture the training dynamics of the projection)
                        self.svals[proj_idx] = self.svals[proj_idx] + [svals]

                    svd_fig = gen_plot(self.svals)

                    self.writer.add_figure('SingularValues', svd_fig, global_step=self.get_iter(runner))


    @master_only
    def after_run(self, runner):
        self.writer.close()