# mobilenetv2/efficientnet-b0 + decoder implementation
# Model architecture based on DenseDepth but implementation is from https://github.com/alinstein/Depth_estimation/blob/master/Mobile_model.py

from collections import OrderedDict
import torch
import torch.nn as nn
import torch.nn.functional as F
from .utils import MODELS
from mmcv.runner import BaseModule, auto_fp16
import torch.distributed as dist
from torchvision.models import mobilenet_v2, MobileNet_V2_Weights, efficientnet_b0, EfficientNet_B0_Weights
from timm.models.vision_transformer import trunc_normal_


class UpSample(nn.Sequential):
    def __init__(self, skip_input, output_features, depthwise_layer_2=False):
        super(UpSample, self).__init__()
        self.convA = nn.Conv2d(skip_input, output_features, kernel_size=3, stride=1, padding=1)
        self.leakyreluA = nn.LeakyReLU(0.2)
        self.convB = nn.Conv2d(output_features, output_features, kernel_size=3, stride=1, padding=1, groups=output_features if depthwise_layer_2 else 1)
        self.leakyreluB = nn.LeakyReLU(0.2)

    def forward(self, x, concat_with):
        up_x = F.interpolate(x, size=[concat_with.size(2), concat_with.size(3)], mode='bilinear', align_corners=True)
        return self.leakyreluB( self.convB( self.leakyreluA(self.convA( torch.cat([up_x, concat_with], dim=1) ) ) )  )


class UpSample_1layer(nn.Sequential):
    def __init__(self, input_features, output_features):
        super().__init__()
        self.convA = nn.Conv2d(input_features, output_features, kernel_size=3, stride=1, padding=1)
        self.leakyreluA = nn.LeakyReLU(0.2)

    def forward(self, x, concat_with):
        up_x = F.interpolate(x, size=[concat_with.size(2), concat_with.size(3)], mode='bilinear', align_corners=True)
        return self.leakyreluA(self.convA( torch.cat([up_x, concat_with], dim=1) ) )


class ULightDecoder_skip_4b(nn.Module):
    """ ULight = UltraLight, 4b = 4 blocks """
    def __init__(self, num_features=1280, decoder_width = .6, encoder='mobilenetv2'):
        super().__init__()
        self.num_features = num_features
        self.features = int(self.num_features * decoder_width)

        self.conv2 = nn.Conv2d(self.num_features, self.features, kernel_size=1, stride=1, padding=0)    # NB: Padding changed to 0 (compared to normal UpSample)
        if encoder == 'efficientnet-b0':
            self.up0 = UpSample_1layer(input_features=self.features//1 + 320, output_features=self.features//2)
            # self.up1 = UpSample_1layer(input_features=self.features//2 + 192, output_features=self.features//2)
            self.up2 = UpSample_1layer(input_features=self.features//2 + 80, output_features=self.features//4)
            self.up3 = UpSample_1layer(input_features=self.features//4 + 40, output_features=self.features//8)
            # self.up4 = UpSample_1layer(input_features=self.features//8 + 24, output_features=self.features//8)
            self.up5 = UpSample_1layer(input_features=self.features//8 + 16, output_features=self.features//16)
            self.conv3 = nn.Conv2d(self.features//16, 1, kernel_size=3, stride=1, padding=1)
        else:
            self.up0 = UpSample_1layer(input_features=self.features//1 + 320, output_features=self.features//2)
            # self.up1 = UpSample_1layer(input_features=self.features//2 + 160, output_features=self.features//2)
            self.up2 = UpSample_1layer(input_features=self.features//2 + 64, output_features=self.features//4)
            self.up3 = UpSample_1layer(input_features=self.features//4 + 32, output_features=self.features//8)
            # self.up4 = UpSample_1layer(input_features=self.features//8 + 24, output_features=self.features//8)
            self.up5 = UpSample_1layer(input_features=self.features//8 + 16, output_features=self.features//16)

            self.conv3 = nn.Conv2d(self.features//16, 1, kernel_size=3, stride=1, padding=1)

    def forward(self, features):
        x_block0, x_block1, x_block2, x_block3, x_block4, x_block5, x_block6 = features[2], features[4], features[6], features[9], features[15],features[18],features[19]
    
        x_d0 = self.conv2(x_block6)
        x_d1 = self.up0(x_d0, x_block5)
        # x_d2 = self.up1(x_d1, x_block4)
        x_d3 = self.up2(x_d1, x_block3)
        x_d4 = self.up3(x_d3, x_block2)
        # x_d5 = self.up4(x_d4, x_block1)
        x_d6 = self.up5(x_d4, x_block0)
        return self.conv3(x_d6)


class Decoder(nn.Module):
    def __init__(self, num_features=1280, decoder_width = .6):
        super(Decoder, self).__init__()
        self.num_features = num_features
        self.features = int(self.num_features * decoder_width)

        self.conv2 = nn.Conv2d(self.num_features, self.features, kernel_size=1, stride=1, padding=1)
        
        self.up0 = UpSample(skip_input=self.features//1 + 320, output_features=self.features//2)
        self.up1 = UpSample(skip_input=self.features//2 + 160, output_features=self.features//2)
        self.up2 = UpSample(skip_input=self.features//2 + 64, output_features=self.features//4)
        self.up3 = UpSample(skip_input=self.features//4 + 32, output_features=self.features//8)
        self.up4 = UpSample(skip_input=self.features//8 +  24, output_features=self.features//8)
        self.up5 = UpSample(skip_input=self.features//8 +  16, output_features=self.features//16)

        self.conv3 = nn.Conv2d(self.features//16, 1, kernel_size=3, stride=1, padding=1)

    def forward(self, features):
        x_block0, x_block1, x_block2, x_block3, x_block4, x_block5, x_block6 = features[2], features[4], features[6], features[9], features[15],features[18],features[19]
        x_d0 = self.conv2(x_block6)
        x_d1 = self.up0(x_d0, x_block5)
        x_d2 = self.up1(x_d1, x_block4)
        x_d3 = self.up2(x_d2, x_block3)
        x_d4 = self.up3(x_d3, x_block2)
        x_d5 = self.up4(x_d4, x_block1)
        x_d6 = self.up5(x_d5, x_block0)
        return self.conv3(x_d6)


class Decoder_dl2(nn.Module):
    def __init__(self, num_features=1280, decoder_width = .6):
        super().__init__()
        self.num_features = num_features
        self.features = int(self.num_features * decoder_width)

        self.conv2 = nn.Conv2d(self.num_features, self.features, kernel_size=1, stride=1, padding=1)
        
        self.up0 = UpSample(skip_input=self.features//1 + 320, output_features=self.features//2, depthwise_layer_2=True)
        self.up1 = UpSample(skip_input=self.features//2 + 160, output_features=self.features//2, depthwise_layer_2=True)
        self.up2 = UpSample(skip_input=self.features//2 + 64, output_features=self.features//4, depthwise_layer_2=True)
        self.up3 = UpSample(skip_input=self.features//4 + 32, output_features=self.features//8, depthwise_layer_2=True)
        self.up4 = UpSample(skip_input=self.features//8 +  24, output_features=self.features//8, depthwise_layer_2=True)
        self.up5 = UpSample(skip_input=self.features//8 +  16, output_features=self.features//16, depthwise_layer_2=True)

        self.conv3 = nn.Conv2d(self.features//16, 1, kernel_size=3, stride=1, padding=1)

    def forward(self, features):
        x_block0, x_block1, x_block2, x_block3, x_block4, x_block5, x_block6 = features[2], features[4], features[6], features[9], features[15],features[18],features[19]
        x_d0 = self.conv2(x_block6)
        x_d1 = self.up0(x_d0, x_block5)
        x_d2 = self.up1(x_d1, x_block4)
        x_d3 = self.up2(x_d2, x_block3)
        x_d4 = self.up3(x_d3, x_block2)
        x_d5 = self.up4(x_d4, x_block1)
        x_d6 = self.up5(x_d5, x_block0)
        return self.conv3(x_d6)


class Encoder(nn.Module):
    def __init__(self, mode=None, pretrained=True):
        super(Encoder, self).__init__()
        self.mode = mode
        self.pretrained = pretrained
        match self.mode:
            case "efficientnet-b0":
                if self.pretrained:
                    weights = EfficientNet_B0_Weights.IMAGENET1K_V1
                else:
                    weights = None
                self.original_model = efficientnet_b0(weights=weights)
                del self.original_model.classifier  # Remove classifier head
            case _: # "mobilenetv2"
                if self.pretrained:
                    weights = MobileNet_V2_Weights.IMAGENET1K_V2
                else:
                    weights = None
                self.original_model = mobilenet_v2(weights=weights)
                del self.original_model.classifier  # Remove classifier head

    def forward(self, x):
        features = [x]
        for k, v in self.original_model.features._modules.items(): features.append( v(features[-1]) )
        if self.mode == "efficientnet-b0":
            new_features = [features[0]]*20
            new_features[2] = features[2]
            new_features[4] = features[3]
            new_features[6] = features[4]
            new_features[9] = features[5]
            new_features[15] = features[7]
            new_features[18] = features[8]
            new_features[19] = features[9]
            features = new_features
        return features


@MODELS.register_module()
class GenericEncDec(BaseModule):
    def __init__(self, encoder_mode='mobilenetv2', decoder_mode=None, pretrained_encoder=True):
        super().__init__()
        self.encoder = Encoder(mode=encoder_mode, pretrained=pretrained_encoder)
        self.decoder_mode = decoder_mode
        if self.decoder_mode == "decoder_dl2":
            self.decoder = Decoder_dl2()
        elif self.decoder_mode == "ulight_decoder_skip_4b":
            self.decoder = ULightDecoder_skip_4b(encoder=encoder_mode)
        else:
            self.decoder = Decoder()


    def forward(self, x):
        return self.decoder(self.encoder(x))
    




