# TwoModelWrapper.py
# Defines a generic two-model wrapper with configurable routing
# Should use configurable task heads with their own losses
# Aims to implement the train_step etc. methods required by the experiment runner.

import torch

from .GenericEncDec import GenericEncDec

from .utils import MODELS
from mmcv.runner import BaseModule, auto_fp16
from mmdet.core import bbox2result
from .utils import build_model
from collections import OrderedDict
import torch.distributed as dist
import matplotlib.pyplot as plt
import torch.nn as nn
import torch.nn.functional as F
from operator import itemgetter
import mmcv
import numpy as np
from mmdet.core.visualization import imshow_det_bboxes
from timm.models.vision_transformer import trunc_normal_
from mmcv.runner import load_checkpoint
from .depth.depth_head import SiLogLoss

from .ait import AiT, Projection
from .vit import ViT
import torchvision

@MODELS.register_module()
class TwoModelWrapper(BaseModule):
    def __init__(self,
                 model_1=None,
                 model_2=None,
                 projection_arch=None,
                 freeze_model_1=False,
                 freeze_model_2=False,
                 freeze_projection=False,
                 projection_metric=None,
                 model_part_names=dict(),
                 projection_direction="T->S",  # Either S->T or T->S
                 init_cfg=None,
                 resizeto=0,
                 padto=0,
                 mode=None, # If undefined, will run model_1 end to end, with no distillation
                 model_1_load_from=None,
                 model_2_load_from=None,
                 projection_load_from=None,
                 random_model_1_input=False,
                 **kwargs
                ):
        super().__init__(init_cfg)

        self.model_part_names=model_part_names
        self.resizeto = resizeto
        self.padto = padto
        self.mode = mode
        self.freeze_model_1 = freeze_model_1
        self.freeze_model_2 = freeze_model_2
        self.freeze_projection = freeze_projection
        self.model_1_load_from = model_1_load_from
        self.model_2_load_from = model_2_load_from
        self.projection_load_from = [] if projection_load_from is None else projection_load_from
        self.projection_arch = projection_arch
        self.projection_metric = projection_metric
        self.random_model_1_input = random_model_1_input

        # for ViT, we just center crop
        self.center_crop = torchvision.transforms.CenterCrop((224, 224))

        assert model_1 is not None, "Must provide model_1 config."
        self.model_1 = build_model(model_1)

        self.model_2 = None
        if model_2 is not None:
            assert self.mode in ["distillation","enc-proj-dec"], "Error: Two models defined but not in distillation mode"
            self.model_2 = build_model(model_2)
        
        self.projections = nn.ModuleList([])
        if self.mode in ["distillation"]:
            assert isinstance(self.model_2, GenericEncDec), "Not implemented for non-MNV2 or EffNet students"

            # teacher: model_1, student: model_2
            # we want to project from student (in_feats) to teacher (out_feats)

            if isinstance(self.model_1, AiT):
                t_feats = self.model_1.transformer.in_chans
            elif isinstance(self.model_1, ViT):
                t_feats = self.model_1.backbone.hidden_dim
            elif isinstance(self.model_1, GenericEncDec):
                t_feats = self.model_1.decoder.num_features

            s_feats = self.model_2.decoder.num_features

            self.projection_direction = projection_direction
            if projection_direction == "S->T":
                in_feats = [s_feats]
                out_feats = [t_feats]
            elif projection_direction == "T->S":
                in_feats = [t_feats]
                out_feats = [s_feats]
            else:
                raise NotImplementedError
            
            for in_feat, out_feat in zip(in_feats, out_feats):
                projection_cfg = {
                    "type": Projection, 
                    "mode": self.projection_arch, 
                    "in_feats": in_feat,
                    "out_feats": out_feat,
                }
                projection = build_model(projection_cfg)
                self.projections.append(projection)
        
        elif self.mode in ["enc-proj-dec"]:
            assert isinstance(self.model_1, AiT), "Not implemented for non-AiT teachers"
            assert isinstance(self.model_2, GenericEncDec), "Not implemented for non-MNV2 students"
            in_feats = [128, 256, 512, 1024]
            out_feats = [16, 24, 32, 64, 160, 320, 1280]
            projection_cfg0 = {
                "type": Projection, 
                "mode": self.projection_arch, 
                "in_feats": in_feats[0],
                "out_feats": out_feats[0],
            }
            projection_cfg1 = {
                "type": Projection, 
                "mode": self.projection_arch, 
                "in_feats": in_feats[0],
                "out_feats": out_feats[1],
            }
            projection_cfg2 = {
                "type": Projection,
                "mode": self.projection_arch,
                "in_feats": in_feats[1],
                "out_feats": out_feats[2],
            }
            projection_cfg3 = {
                "type": Projection,
                "mode": self.projection_arch,
                "in_feats": in_feats[2],
                "out_feats": out_feats[3],
            }
            projection_cfg4 = {
                "type": Projection,
                "mode": self.projection_arch,
                "in_feats": in_feats[3],
                "out_feats": out_feats[4],
            }
            projection_cfg5 = {
                "type": Projection,
                "mode": self.projection_arch,
                "in_feats": in_feats[3],
                "out_feats": out_feats[5],
            }
            projection_cfg6 = {
                "type": Projection,
                "mode": self.projection_arch,
                "in_feats": in_feats[3],
                "out_feats": out_feats[6],
            }
            self.projection0 = build_model(projection_cfg0)
            self.projection1 = build_model(projection_cfg1)
            self.projection2 = build_model(projection_cfg2)
            self.projection3 = build_model(projection_cfg3)
            self.projection4 = build_model(projection_cfg4)
            self.projection5 = build_model(projection_cfg5)
            self.projection6 = build_model(projection_cfg6)

        self.depth_loss = SiLogLoss()

        # metric is some state-less function measuring the distance between vector
        if projection_metric in ["l2", "mse", "L2"]:
            self.metric_fn = lambda z_s, z_t: F.mse_loss(z_s, z_t)
        else:
            if len(self.projections) > 0:
                raise Exception("Are you sure?")
            self.metric_fn = lambda z_s, z_t: torch.tensor(0., device=f'cuda:{torch.cuda.current_device()}')

        return


    def maybe_freeze_models(self):
        """ Put model_1, model_2, and projection in the correct train/eval
        mode depending on freeze arguments. 
        """
        if self.freeze_model_1:
            for param in self.model_1.parameters():
                param.requires_grad = False
            self.model_1.eval()
        elif self.training:
            for param in self.model_1.parameters():
                param.requires_grad = True
            self.model_1.train()

        if self.model_2 is not None:
            if self.freeze_model_2:
                for param in self.model_2.parameters():
                    param.requires_grad = False
                self.model_2.eval()
            elif self.training:
                for param in self.model_2.parameters():
                    param.requires_grad = True
                self.model_2.train()
        
        if len(self.projections) > 0:
            if self.freeze_projection:
                for projection in self.projections:
                    for param in projection.parameters():
                        param.requires_grad = False
                    projection.eval()
            elif self.training:
                for projection in self.projections:
                    for param in projection.parameters():
                        param.requires_grad = True
                    projection.train()
                

    def init_weights(self):
        self.model_1.init_weights()
        if self.model_2 is not None:
            self.model_2.init_weights()

        if self.mode in ["enc-proj-dec"]:
            self.projection0.init_weights()
            self.projection1.init_weights()
            self.projection2.init_weights()
            self.projection3.init_weights()
            self.projection4.init_weights()
            self.projection5.init_weights()
            self.projection6.init_weights()

        if len(self.projections) > 0:
            for i in range(len(self.projections)):
                self.projections[i].init_weights()

        self.maybe_freeze_models()

        if self.model_1_load_from is not None:
            if isinstance(self.model_1, ViT):
                load_checkpoint(self.model_1.backbone, self.model_1_load_from)
            else:
                load_checkpoint(self.model_1, self.model_1_load_from)

        if self.mode in ["enc-proj-dec"]:
            if self.model_2_load_from is not None and self.model_2 is not None:
                # convert keys from model_1 to no prefix
                new_dict = self.convert_keys(torch.load(self.model_2_load_from)['state_dict'])
                self.model_2.load_state_dict(new_dict)
                # load_checkpoint(self.model_2, self.model_2_load_from)
            if self.projection_load_from is not None and self.projection is not None:
                load_checkpoint(self.projection, self.projection_load_from) 
        else:
            # Ours
            if self.model_2_load_from is not None and self.model_2 is not None:
                load_checkpoint(self.model_2, self.model_2_load_from)

            if isinstance(self.projection_load_from, str):
                # backward compatible to when we exclusively had one projector
                self.projection_load_from = [self.projection_load_from]

            if len(self.projection_load_from) > 0 and len(self.projections) > 0:
                for i, (_, projection_load_from) in enumerate(zip(self.projections, self.projection_load_from)):
                    load_checkpoint(self.projections[i], projection_load_from)    

        if isinstance(self.model_1, AiT):
            del self.model_1.transformer
            if hasattr(self.model_1, 'det_head'):
                del self.model_1.det_head
            if hasattr(self.model_1, 'insseg_head'):
                del self.model_1.insseg_head
            if hasattr(self.model_1, 'depth_head'):
                del self.model_1.depth_head
        elif isinstance(self.model_1, GenericEncDec) and self.model_2 is not None:
            del self.model_1.decoder


    def convert_keys(self, pretrained_model_dict):
        new_dict = {}
        for k, v in pretrained_model_dict.items():
            new_key = k.replace('model_1.', '')
            new_dict[new_key] = v
        return new_dict


    def forward_train(self, img, **kwargs):
        self.maybe_freeze_models()

        img = torch.cat([self.resize(i) for i in img], dim=0)
        gt = kwargs['gt']

        losses = {
            'det_loss': torch.tensor(0., device=f'cuda:{torch.cuda.current_device()}'), 'det_cls_ce': torch.tensor(0., device=f'cuda:{torch.cuda.current_device()}'), 'det_bbox_ce': torch.tensor(0., device=f'cuda:{torch.cuda.current_device()}'),
            'insseg_loss': torch.tensor(0., device=f'cuda:{torch.cuda.current_device()}'), 'insseg_cls_ce': torch.tensor(0., device=f'cuda:{torch.cuda.current_device()}'), 'insseg_bbox_ce': torch.tensor(0., device=f'cuda:{torch.cuda.current_device()}'), 'insseg_mask_ce': torch.tensor(0., device=f'cuda:{torch.cuda.current_device()}'), 'insseg_decoder_aux_ce': torch.tensor(0., device=f'cuda:{torch.cuda.current_device()}'),
            'depth_loss': torch.tensor(0., device=f'cuda:{torch.cuda.current_device()}'), 'depth_ce': torch.tensor(0., device=f'cuda:{torch.cuda.current_device()}'), 'depth_decoder_aux_ce': torch.tensor(0., device=f'cuda:{torch.cuda.current_device()}'),
        }
        if isinstance(img, list):
            assert len(img) == 1, "Too many images fed in, is this meant to be multi-task?"
            img = img[0]
        if isinstance(gt, list):
            assert len(gt) == 1, "Too many gts fed in, is this meant to be multi-task?"
            gt = gt[0]

        if self.random_model_1_input:
            img_m1 = torch.randn_like(img)  # torch.randn_like gives normal dist. with mean 0, var 1.
        else:
            img_m1 = img

        # getattr() doesn't produce nonzero gradients, __getattr__ seems to work but I'm not certain so just using ifs
        # m1_feats = self.model_1.__getattr__(self.model_part_names.get("m1_encoder", "encoder"))(img)

        if isinstance(self.model_1, AiT):
            m1_feats = self.model_1.backbone(img_m1)
            if self.mode not in ["enc-proj-dec",]:
                m1_feats = m1_feats[0]
        elif isinstance(self.model_1, GenericEncDec):
            if self.model_2 is not None:
                m1_feats = self.model_1.encoder(img_m1)[-1] # Only take output dim; the rest is intermediate features.
            else:
                m1_feats = self.model_1.encoder(img_m1) 
        elif isinstance(self.model_1, ViT):
            img_crop = self.center_crop(img_m1)
            m1_feats = self.model_1.forward_feats(img_crop)
            m1_feats = self.model_1.extract_hw_feats(m1_feats)
        else:
            raise NotImplementedError

        if self.mode in ["distillation"]:
            if isinstance(self.model_2, AiT):
                m2_feats = self.model_2.backbone(img)
            elif isinstance(self.model_2, GenericEncDec):
                m2_feats = self.model_2.encoder(img)
            else:
                raise NotImplementedError

            if self.projection_direction == "S->T":
                z_s = [self.projections[0](m2_feats[-1])]
                z_t = [m1_feats]
            elif self.projection_direction == "T->S":
                z_s = [self.projections[0](m1_feats)]
                z_t = [m2_feats[-1]]
            else:
                raise NotImplementedError

            if isinstance(self.model_1, ViT):
                # pool over HW of both representations for simplicity T: (14, 14), S: (15, 20)
                # this is a valid distillation method (I have seen this proposed in some papers)
                for i, (x, y) in enumerate(zip(z_s, z_t)):
                    b, c, h, w = x.shape
                    z_s[i] = x.view(b, c, h * w).mean(2).view(b, c, 1, 1)

                    h, w = y.shape[2:]
                    z_t[i] = y.view(b, c, h * w).mean(2).view(b, c, 1, 1)
                
            losses['projection_loss'] = 0.0

            assert len(z_s) == len(z_t) == 1
            loss_weightings = [1]
            for x, y, weight in zip(z_s, z_t, loss_weightings):
                losses['projection_loss'] += weight * self.metric_fn(x, y)

            if isinstance(self.model_2, AiT):
                raise NotImplementedError
                output_pred = self.model_2.depth_head(m2_feats)
            elif isinstance(self.model_2, GenericEncDec):
                output_pred = self.model_2.decoder(m2_feats)
            else:
                raise NotImplementedError
            
        elif self.mode in ["enc-proj-dec",]:
            proj_feats0 = self.projection0(m1_feats[0])
            proj_feats1 = self.projection1(m1_feats[0])
            proj_feats2 = self.projection2(m1_feats[1])
            proj_feats3 = self.projection3(m1_feats[2])
            proj_feats4 = self.projection4(m1_feats[3])
            proj_feats5 = self.projection5(m1_feats[3])
            proj_feats6 = self.projection6(m1_feats[3])
            # list of length 20
            proj_feats = [proj_feats0 for _ in range(20)]
            proj_feats[2] = torch.nn.functional.interpolate(proj_feats0, size=(240, 320), mode='bilinear', align_corners=False)
            proj_feats[4] = proj_feats1
            proj_feats[6] = proj_feats2
            proj_feats[9] = proj_feats3
            proj_feats[15] = proj_feats4
            proj_feats[18] = proj_feats5
            proj_feats[19] = proj_feats6

            output_pred = getattr(self.model_2, self.model_part_names.get("m2_decoder", "decoder"))(proj_feats)

        elif self.mode in [None,]:
            # output_pred = getattr(self.model_1, self.model_part_names.get("m1_decoder", "decoder"))(m1_feats)
            output_pred = self.model_1.decoder(m1_feats)
        else:
            raise NotImplementedError

        # Final resizing: needed because some backbones output half resolution.
        if output_pred.shape[-2:] != img.shape[-2:]:
            output_pred = F.interpolate(output_pred, (img.shape[-2], img.shape[-1]), mode='bilinear')

        losses['depth_loss'] = self.depth_loss(output_pred, gt['depth'])

        return losses


    def forward_test(self, img, img_metas=None, task='depth', **kwargs):
        """
        Args:
            imgs (List[Tensor]): the outer list indicates test-time
                augmentations and inner Tensor should have a shape NxCxHxW,
                which contains all images in the batch.
            img_metas (List[List[dict]]): the outer list indicates test-time
                augs (multiscale, flip, etc.) and the inner list indicates
                images in a batch.
        """
        if task in ['det', 'insseg']:
            raise NotImplementedError
            for var, name in [(imgs, 'imgs'), (img_metas, 'img_metas')]:
                if not isinstance(var, list):
                    raise TypeError(
                        f'{name} must be a list, but got {type(var)}')

            num_augs = len(imgs)
            if num_augs != len(img_metas):
                raise ValueError(f'num of augmentations ({len(imgs)}) '
                                 f'!= num of image meta ({len(img_metas)})')

            for img, img_meta in zip(imgs, img_metas):
                batch_size = len(img_meta)
                for img_id in range(batch_size):
                    img_meta[img_id]['batch_input_shape'] = tuple(
                        img.size()[-2:])

            if num_augs == 1:
                if 'proposals' in kwargs:
                    kwargs['proposals'] = kwargs['proposals'][0]
                if task == 'det':
                    return self.simple_test_box(imgs[0], img_metas[0], **kwargs)
                else:
                    return self.simple_test_insseg(imgs[0], img_metas[0], **kwargs)
            else:
                raise NotImplementedError
        elif task == 'depth':
            depth_gt = kwargs['depth']
            depth_gt = depth_gt.squeeze(1)

            if self.mode in ["distillation",]:
                m2_feats = getattr(self.model_2, self.model_part_names.get("m2_encoder", "encoder"))(img)
                output_pred = getattr(self.model_2, self.model_part_names.get("m1_decoder", "decoder"))(m2_feats)
            elif self.mode in ["enc-proj-dec",]:
                m1_feats = getattr(self.model_1, self.model_part_names.get("m1_encoder", "encoder"))(img)
                proj_feats0 = self.projection0(m1_feats[0])
                proj_feats1 = self.projection1(m1_feats[0])
                proj_feats2 = self.projection2(m1_feats[1])
                proj_feats3 = self.projection3(m1_feats[2])
                proj_feats4 = self.projection4(m1_feats[3])
                proj_feats5 = self.projection5(m1_feats[3])
                proj_feats6 = self.projection6(m1_feats[3])
                # list of length 20
                proj_feats = [proj_feats0 for _ in range(20)]
                proj_feats[2] = torch.nn.functional.interpolate(proj_feats0, size=(240, 320), mode='bilinear', align_corners=False)
                proj_feats[4] = proj_feats1
                proj_feats[6] = proj_feats2
                proj_feats[9] = proj_feats3
                proj_feats[15] = proj_feats4
                proj_feats[18] = proj_feats5
                proj_feats[19] = proj_feats6
                output_pred = getattr(self.model_2, self.model_part_names.get("m2_decoder", "decoder"))(proj_feats)
            elif self.mode in [None,]:
                m1_feats = self.model_1.encoder(img)
                output_pred = self.model_1.decoder(m1_feats)

            if output_pred.shape[-2:] != img.shape[-2:]:
                output_pred = F.interpolate(output_pred, (img.shape[-2], img.shape[-1]), mode='bilinear')

            return list(zip(output_pred.squeeze(1).clip(min=1e-3).detach().cpu(), depth_gt.cpu()))

        else:
            raise NotImplementedError

    def resize(self, img):
        _, _, H, W = img.shape
        if self.resizeto > 0:
            img = F.interpolate(img, self.resizeto, mode='bilinear')
        if self.padto > 0:
            img = F.pad(img, (0, self.padto-W, 0, self.padto-H))
        return img

    @auto_fp16(apply_to=('img', ))
    def forward(self, img, return_loss=True, **kwargs):
        if return_loss:
            return self.forward_train(img, **kwargs)
        else:
            return self.forward_test(img, **kwargs)

    def simple_test_box(self, img, img_metas, rescale=False):
        feat = self.backbone(img)
        results_list = self.det_head.simple_test(
            feat, img_metas, rescale=rescale)
        bbox_results = [
            bbox2result(det_bboxes, det_labels, self.det_head.num_classes)
            for det_bboxes, det_labels in results_list
        ]
        return bbox_results

    def simple_test_insseg(self, img, img_metas, rescale=False):
        """Test function without test-time augmentation.

        Args:
            img (torch.Tensor): Images with shape (N, C, H, W).
            img_metas (list[dict]): List of image information.
            rescale (bool, optional): Whether to rescale the results.
                Defaults to False.

        Returns:
            list[list[np.ndarray]]: BBox results of each image and classes.
                The outer list corresponds to each image. The inner list
                corresponds to each class.
        """
        feat = self.backbone(img)
        results_list = self.insseg_head.simple_test(
            feat, img_metas, rescale=rescale)
        bbox_results = [
            bbox2result(det_bboxes, det_labels, self.insseg_head.num_classes)
            for det_bboxes, det_labels, _ in results_list
        ]
        mask_results = [
            mask2result(det_masks, det_labels, self.insseg_head.num_classes)
            for _, det_labels, det_masks in results_list
        ]
        return list(zip(bbox_results, mask_results))

    def train_step(self, data, optimizer):
        """The iteration step during training.

        This method defines an iteration step during training, except for the
        back propagation and optimizer updating, which are done in an optimizer
        hook. Note that in some complicated cases or models, the whole process
        including back propagation and optimizer updating is also defined in
        this method, such as GAN.

        Args:
            data (dict): The output of dataloader.
            optimizer (:obj:`torch.optim.Optimizer` | dict): The optimizer of
                runner is passed to ``train_step()``. This argument is unused
                and reserved.

        Returns:
            dict: It should contain at least 3 keys: ``loss``, ``log_vars``, \
                ``num_samples``.

                - ``loss`` is a tensor for back propagation, which can be a
                  weighted sum of multiple losses.
                - ``log_vars`` contains all the variables to be sent to the
                  logger.
                - ``num_samples`` indicates the batch size (when the model is
                  DDP, it means the batch size on each GPU), which is used for
                  averaging the logs.
        """
        losses = self(**data)
        loss, log_vars = self._parse_losses(losses)

        if isinstance(data['img'], torch.Tensor):
            num_samples = data['img'].shape[0]
        else:
            num_samples = sum(map(lambda x: x.shape[0], data['img']))
        outputs = dict(
            loss=loss, log_vars=log_vars, num_samples=num_samples)

        return outputs


    def _parse_losses(self, losses):
        """Parse the raw outputs (losses) of the network.

        Args:
            losses (dict): Raw output of the network, which usually contain
                losses and other necessary information.

        Returns:
            tuple[Tensor, dict]: (loss, log_vars), loss is the loss tensor \
                which may be a weighted sum of all losses, log_vars contains \
                all the variables to be sent to the logger.
        """
        log_vars = OrderedDict()
        for loss_name, loss_value in losses.items():
            if isinstance(loss_value, torch.Tensor):
                log_vars[loss_name] = loss_value.mean()
            elif isinstance(loss_value, list):
                log_vars[loss_name] = sum(_loss.mean() for _loss in loss_value)
            else:
                raise TypeError(
                    f'{loss_name} is not a tensor or list of tensors')

        loss = sum(_value for _key, _value in log_vars.items()
                   if 'loss' in _key)

        # If the loss_vars has different length, GPUs will wait infinitely
        if dist.is_available() and dist.is_initialized():
            log_var_length = torch.tensor(len(log_vars), device=loss.device)
            dist.all_reduce(log_var_length)
            message = (f'rank {dist.get_rank()}' +
                       f' len(log_vars): {len(log_vars)}' + ' keys: ' +
                       ','.join(log_vars.keys()))
            assert log_var_length == len(log_vars) * dist.get_world_size(), \
                'loss log variables are different across GPUs!\n' + message

        log_vars['loss'] = loss
        for loss_name, loss_value in log_vars.items():
            # reduce loss when distributed training
            if dist.is_available() and dist.is_initialized():
                loss_value = loss_value.data.clone()
                dist.all_reduce(loss_value.div_(dist.get_world_size()))
            log_vars[loss_name] = loss_value.item()

        return loss, log_vars

    def show_insseg_result(self,
                           img,
                           result,
                           score_thr=0.3,
                           bbox_color=(72, 101, 241),
                           text_color=(72, 101, 241),
                           mask_color=None,
                           thickness=2,
                           font_size=13,
                           win_name='',
                           show=False,
                           wait_time=0,
                           class_names=None,
                           out_file=None):
        """Draw `result` over `img`.
        Args:
            img (str or Tensor): The image to be displayed.
            result (Tensor or tuple): The results to draw over `img`
                bbox_result or (bbox_result, segm_result).
            score_thr (float, optional): Minimum score of bboxes to be shown.
                Default: 0.3.
            bbox_color (str or tuple(int) or :obj:`Color`):Color of bbox lines.
               The tuple of color should be in BGR order. Default: 'green'
            text_color (str or tuple(int) or :obj:`Color`):Color of texts.
               The tuple of color should be in BGR order. Default: 'green'
            mask_color (None or str or tuple(int) or :obj:`Color`):
               Color of masks. The tuple of color should be in BGR order.
               Default: None
            thickness (int): Thickness of lines. Default: 2
            font_size (int): Font size of texts. Default: 13
            win_name (str): The window name. Default: ''
            wait_time (float): Value of waitKey param.
                Default: 0.
            show (bool): Whether to show the image.
                Default: False.
            out_file (str or None): The filename to write the image.
                Default: None.
        Returns:
            img (Tensor): Only if not `show` or `out_file`
        """
        img = mmcv.imread(img)
        img = img.copy()
        if isinstance(result, tuple):
            bbox_result, segm_result = result
            if isinstance(segm_result, tuple):
                segm_result = segm_result[0]  # ms rcnn
        else:
            bbox_result, segm_result = result, None
        bboxes = np.vstack(bbox_result)
        labels = [
            np.full(bbox.shape[0], i, dtype=np.int32)
            for i, bbox in enumerate(bbox_result)
        ]
        labels = np.concatenate(labels)
        # draw segmentation masks
        segms = None
        if segm_result is not None and len(labels) > 0:  # non empty
            segms = mmcv.concat_list(segm_result)
            if isinstance(segms[0], torch.Tensor):
                segms = torch.stack(segms, dim=0).detach().cpu().numpy()
            else:
                segms = np.stack(segms, axis=0)
        # if out_file specified, do not show image in window
        if out_file is not None:
            show = False
        # draw bounding boxes
        img = imshow_det_bboxes(
            img,
            bboxes,
            labels,
            segms,
            class_names=class_names,
            score_thr=score_thr,
            bbox_color=bbox_color,
            text_color=text_color,
            mask_color=mask_color,
            thickness=thickness,
            font_size=font_size,
            win_name=win_name,
            show=show,
            wait_time=wait_time,
            out_file=out_file)

        if not (show or out_file):
            return img

        

def mask2result(masks, labels, num_classes):
    """Convert detection results to a list of numpy arrays.

    Args:
        bboxes (torch.Tensor | np.ndarray): shape (n, 5)
        labels (torch.Tensor | np.ndarray): shape (n, )
        num_classes (int): class number, including background class

    Returns:
        list(ndarray): bbox results of each class
    """
    cls_segms = [[] for _ in range(num_classes)
                 ]  # BG is not included in num_classes
    N = masks.shape[0]
    for i in range(N):
        cls_segms[labels[i]].append(masks[i].detach().cpu().numpy())
    return cls_segms