from collections import OrderedDict
import torch
import torch.nn as nn
import torch.nn.functional as F
from .utils import MODELS
from mmcv.runner import BaseModule, auto_fp16
from timm.models.vision_transformer import trunc_normal_
import torchvision

@MODELS.register_module()
class ViT(BaseModule):
    def __init__(self):
        super().__init__()
        self.backbone = torchvision.models.vit_b_16()

    def _process_input(self, x):
        n, c, h, w = x.shape
        p = self.backbone.patch_size
        torch._assert(h == self.backbone.image_size, f"Wrong image height! Expected {self.backbone.image_size} but got {h}!")
        torch._assert(w == self.backbone.image_size, f"Wrong image width! Expected {self.backbone.image_size} but got {w}!")
        n_h = h // p
        n_w = w // p

        # (n, c, h, w) -> (n, hidden_dim, n_h, n_w)
        x = self.backbone.conv_proj(x)
        # (n, hidden_dim, n_h, n_w) -> (n, hidden_dim, (n_h * n_w))
        x = x.reshape(n, self.backbone.hidden_dim, n_h * n_w)

        # (n, hidden_dim, (n_h * n_w)) -> (n, (n_h * n_w), hidden_dim)
        # The self attention layer expects inputs in the format (N, S, E)
        # where S is the source sequence length, N is the batch size, E is the
        # embedding dimension
        x = x.permute(0, 2, 1)

        return x

    def forward_feats(self, x):
        # Reshape and permute the input tensor
        x = self._process_input(x)
        n = x.shape[0]

        # Expand the class token to the full batch
        batch_class_token = self.backbone.class_token.expand(n, -1, -1)
        x = torch.cat([batch_class_token, x], dim=1)

        x = self.backbone.encoder(x)
        return x

    def extract_hw_feats(self, x):
        # extract patch tokens and reshape
        x = x[:, 1:]

        # assume 224x224 imagenet
        bsz = x.shape[0]
        p = self.backbone.patch_size
        n_h = 224 // p
        n_w = 224 // p

        x = x.reshape(bsz, n_h, n_w, -1)
        x = torch.einsum('nhwc->nchw', x)
        
        return x

    def forward(self, x):
        return self.backbone(x)