# Learning to Project for Cross-Task Knowledge Distillation
This is the official code release for the paper "Learning to Project for Cross-Task Knowledge Distillation", submitted to BMVC 2024.
This version of the code has been anonymized. Upon acceptance, the code and checkpoints will be released publicly.


## Structure
The code is written in Python using PyTorch. It is broken into three parts:

1. `DepthEstimation`: Used for monocular depth estimation (MDE) experiments. This code is based on the [official All In Tokens repository](https://github.com/SwinTransformer/AiT/) [[2]](#2) and is built with [MMDetection](https://github.com/open-mmlab/mmdetection).
2. `Colorization_Sat2Map`: Used for image-to-image translation experiments (colorization of black+white images and satellide-to-map translation). This code is a fork of [the PyTorch implementation of CycleGAN and Pix2Pix](https://github.com/junyanz/pytorch-CycleGAN-and-pix2pix) [[3]](#3)[[4]](#4).
3. `SemanticSegmentation`: Used for semantic segmentation tasks. Based on [this repository](https://github.com/yassouali/pytorch-segmentation/).


## Setup
### Environment
A conda environment file (`conda_environment.yml`) is included. Some requirements are pip only, so the conda installation may fail. To set up, install conda on your system and run the following:

```bash
conda env create -f conda_environment.yml
conda activate learn2proj
cd DepthEstimation
python -m pip install -r requirements.txt
```


### Dataset Setup
For depth estimation, we use the NYUv2 indoor depth estimation dataset [[1]](#1). To set up, follow the instructions in [the DepthEstimation README](DepthEstimation/README.md), which is a copy of the AiT repository's README at the time of forking.

For semantic segmentation, the ADE20K Scene Parsing Challenge dataset is required. To download it, run the following:

```bash
cd SemanticSegmentation
mkdir data
wget -O ./data/ADEChallengeData2016.zip http://data.csail.mit.edu/places/ADEchallenge/ADEChallengeData2016.zip
unzip ./data/ADEChallengeData2016.zip -d ./data
rm ./data/ADEChallengeData2016.zip
```

For image-to-image tasks (CycleGan and Pix2Pix for colorization and satellite-to-map), the CMP Facades and Maps datasets are used. Follow the instructions in [the Colorization_Sat2Map README](Colorization_Sat2Map/README.md) to download and set up these datasets.

### Starting Checkpoints (only for training)
Some of the checkpoints used for teacher models must be downloaded manually, while some are sourced from the [torchvision.models](https://pytorch.org/vision/0.15/models.html) model hub.

For depth, checkpoints should be placed in `DepthEstimation/ckpts`. Follow the instructions in [the DepthEstimation README](DepthEstimation/README.md) to download the All In Tokens pretrained weights for SwinV2-B, both for instance segmentation and depth estimation (both of which we use as teachers in different experiments). You may also need to download the VQVAE weights; these are not directly used, but having them allows the model to load checkpoints without manually changing code. This issue will be fixed in the release version.

In addition, the depth student classification teacher is a ViT-b-16 pretrained on ImageNet 1K from `torchvision.models`. The checkpoint must be manually downloaded using:

```bash
cd DepthEstimation/ckpts
wget https://download.pytorch.org/models/vit_b_16-c867db91.pth
```


## Depth
### Evaluation
To run evaluation with a given checkpoint, first `cd DepthEstimation/ait` then use this command:

```bash
torchrun --nnodes=1 --nproc_per_node=1 --rdzv-backend=c10d --rdzv-endpoint=localhost:0 code/train.py --eval /path/to/ckpt.pth
```

In some cases (when there is only one GPU installed in the machine), you may need to run this instead:

```bash
torchrun --nnodes=1 --nproc_per_node=1 --standalone code/train.py --eval /path/to/ckpt.pth
```

Due to file size constraints for the supplementary material, checkpoints will be released with the final version of the code.


### Training
With a given config file defined, `cd DepthEstimation/ait` and then run:

```bash
torchrun --nnodes=1 --nproc_per_node=1 --rdzv-backend=c10d --rdzv-endpoint=localhost:0 code/train.py --config /path/to/config_file.py
```

As with evaluation, if there is only 1 GPU installed in the machine, add `--standalone` and remove the `--rdzv-*` arguments.

In our setup, only single-GPU training was tested (even on systems with multiple GPUs), so we cannot guarantee that training will work in a multi-GPU setup.

The config file names used follow a shorthand:
- The general format is `<teacher_model><teacher_task><student_encoder><student_decoder><projection_direction>.py`
- Efficientnet-B0 students are denoted `enetb0_`
- MobilenetV2 students are denoted with `mnv2_`
- Different decoder architectures are written after the student name and are named as described in the supplementary material
- Teacher architectures, if used, come first:
    - `n_dist_swinv2b_` is the SwinV2-B used in AiT, either pretrained on `insseg` or `depth`, or `rand`-omly initialized.
    - `vit` is the ViT-B-16 pretrained on Imagenet-1K
- Projection direction is `t2s` or `s2t` for teacher-to-student or student-to-teacher respectively.
- Configs for baselines contain the name of a single model only

## Semantic Segmentation
First, run `cd SemanticSegmentation`. The training command is:
```bash
python train.py --config /path/to/config.json
```

At present, evaluation is implemented at the end of training, therefore the easiest way to run evaluation without changing the code is by resuming training with a finished checkpoint (i.e. has trained for the number of epochs specified in the config file):
```bash
python train.py --resume /path/to/checkpoint.pth
```

## Image-to-image translation
The commands for training and evaluation are exactly as described in [the image-to-image README](Colorization_Sat2Map/README.md). The code is currently set up to run CycleGAN with a classification teacher; changing to other tasks is done in the code. The final release of the code will have tasks configurable using command line arguments or config files.


## References
<a id="1">[1]</a> 
Nathan Silberman, Derek Hoiem, Pushmeet Kohli and Rob Fergus, \
Indoor Segmentation and Support Inference from RGBD Images, \
ECCV 2012

<a id="2">[2]</a> 
All in Tokens: Unifying Output Space of Visual Tasks via Soft Token,\
Ning, Jia and Li, Chen and Zhang, Zheng and Geng, Zigang and Dai, Qi and He, Kun and Hu, Han, \
arXiv preprint arXiv:2301.02229, \
2023

<a id="3">[3]</a> 
Unpaired Image-to-Image Translation using Cycle-Consistent Adversarial Networks,\
Zhu, Jun-Yan and Park, Taesung and Isola, Phillip and Efros, Alexei A, \
2017 IEEE International Conference on Computer Vision (ICCV),  \
2017

<a id="4">[4]</a> 
Image-to-Image Translation with Conditional Adversarial Networks,\
Isola, Phillip and Zhu, Jun-Yan and Zhou, Tinghui and Efros, Alexei A,\
2017 IEEE Conference on Computer Vision and Pattern Recognition (CVPR), \
2017
