import os
import json
import argparse
import torch
import dataloaders
import models
import inspect
import math
from utils import losses
from utils import Logger
from utils.torchsummary import summary
from trainer import Trainer
from torchvision.models import resnet50, ResNet50_Weights
from torchvision.models.segmentation import deeplabv3_resnet50, DeepLabV3_ResNet50_Weights
from transformers import AutoImageProcessor, SegformerForSemanticSegmentation, SegformerConfig
import torch.nn as nn

def get_instance(module, name, config, *args):
    # GET THE CORRESPONDING CLASS / FCT 
    return getattr(module, config[name]['type'])(*args, **config[name]['args'])

def main(config, resume):
    train_logger = Logger()

    # DATA LOADERS
    train_loader = get_instance(dataloaders, 'train_loader', config)
    val_loader = get_instance(dataloaders, 'val_loader', config)

    # MODEL
    model = get_instance(models, 'arch', config, train_loader.dataset.num_classes)
    print("========================== Model Info ==========================")

    print(f'Main model (student): {model.__class__}')

    # aux_model = torch.hub.load('NVIDIA/DeepLearningExamples:torchhub', 'nvidia_resnet50', pretrained=True)
    # aux_model = models.FCNResnet50()

    aux_model = None
    projector_t2s = None
    projector_s2t = None

    match config.get('aux_arch'):
        case "rn50":
            aux_model = resnet50(weights=ResNet50_Weights.DEFAULT if config['aux_weights'] == "default" else None)
        case "deeplabv3_rn50":
            aux_model = deeplabv3_resnet50(weights=DeepLabV3_ResNet50_Weights.DEFAULT if config['aux_weights'] == "default" else None)
        case "segformer" | "segformer-b0":
            if config['aux_weights'] == "default":
                aux_model = SegformerForSemanticSegmentation.from_pretrained("nvidia/segformer-b0-finetuned-ade-512-512")
            else:
                aux_model = SegformerForSemanticSegmentation(SegformerConfig())
        case "segformer-b1":
            if config['aux_weights'] == "default":
                aux_model = SegformerForSemanticSegmentation.from_pretrained("nvidia/segformer-b1-finetuned-ade-512-512")
            else:
                # Easier to load the config from the pretrained model and then to re-instantiate a random model using that
                tmp_config = SegformerForSemanticSegmentation.from_pretrained("nvidia/segformer-b1-finetuned-ade-512-512").config
                aux_model = SegformerForSemanticSegmentation(tmp_config)
        case _:
            pass

    if aux_model is not None:
        aux_model.eval()
    match config['arch']['args']['backbone']:
        case "resnet50":
            student_feats = 1024 if config.get("no_last_layer_or_interpolate") else 2048
        case _:
            student_feats = 256 if config.get("no_last_layer_or_interpolate") else 512

    match config.get('aux_arch'):
        case "rn50":
            teacher_feats = 1024 if config.get("no_last_layer_or_interpolate") else 2048
        case "deeplabv3_rn50":
            assert not config.get("no_last_layer_or_interpolate"), "The no last layer is untested here"
            teacher_feats = 2048
        case "segformer" | "segformer-b0":
            assert not config.get("no_last_layer_or_interpolate"), "The no last layer is untested here"
            teacher_feats = 256
        case "segformer-b1":
            assert not config.get("no_last_layer_or_interpolate"), "The no last layer is untested here"
            teacher_feats = 512
        case _:
            pass

    match config['projection_dir']:
        case "t2s":
            projector_t2s = nn.Conv2d(teacher_feats, student_feats, kernel_size=1)
        case "s2t":
            projector_s2t = nn.Conv2d(student_feats, teacher_feats, kernel_size=1)
        case "both":
            projector_s2t = nn.Conv2d(student_feats, teacher_feats, kernel_size=1)
            projector_t2s = nn.Conv2d(teacher_feats, student_feats, kernel_size=1)
        case "ident":
            projector_s2t = nn.Identity()

    print(f'Aux (teacher) model: {aux_model.__class__}')

    print("================================================================")

    # del aux_model.avgpool
    # del aux_model.fc

    # LOSS
    loss = getattr(losses, config['loss'])(ignore_index = config['ignore_index'])

    # TRAINING
    trainer = Trainer(
        model=model,
        aux_model=aux_model,
        projector_s2t=projector_s2t,
        projector_t2s=projector_t2s,
        loss=loss,
        resume=resume,
        config=config,
        train_loader=train_loader,
        val_loader=val_loader,
        train_logger=train_logger)

    trainer.train()

if __name__=='__main__':
    # PARSE THE ARGS
    parser = argparse.ArgumentParser(description='PyTorch Training')
    parser.add_argument('-c', '--config', default='config.json',type=str,
                        help='Path to the config file (default: config.json)')
    parser.add_argument('-r', '--resume', default=None, type=str,
                        help='Path to the .pth model checkpoint to resume training')
    parser.add_argument('-d', '--device', default=None, type=str,
                           help='indices of GPUs to enable (default: all)')
    parser.add_argument('--debug', action="store_true",
                           help='Set train and test batch size to 2 and set num_workers on dataset to 0')

    args = parser.parse_args()

    config = json.load(open(args.config))
    if args.resume:
        config = torch.load(args.resume)['config']
    if args.device:
        os.environ["CUDA_VISIBLE_DEVICES"] = args.device
    
    if args.debug:
        print("========================== DEBUGGING MODE ==========================")
        print("========================== DEBUGGING MODE ==========================")
        print("========================== DEBUGGING MODE ==========================")
        print("========================== DEBUGGING MODE ==========================")
        config['train_loader']['args']['batch_size'] = 2
        # config['train_loader']['args']['num_workers'] = 0
        config['val_loader']['args']['batch_size'] = 2
        # config['val_loader']['args']['num_workers'] = 0

    main(config, args.resume)