import torch
import time
import numpy as np
from torchvision.utils import make_grid
from torchvision import transforms
from utils import transforms as local_transforms
from base import BaseTrainer, DataPrefetcher
from utils.helpers import colorize_mask
from utils.metrics import eval_metrics, AverageMeter
from tqdm import tqdm
import torch.nn.functional as F
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd

class Trainer(BaseTrainer):
    def __init__(self, model, loss, resume, config, train_loader, val_loader=None, train_logger=None, prefetch=True, aux_model=None, projector_s2t=None, projector_t2s=None):
        super(Trainer, self).__init__(model, loss, resume, config, train_loader, val_loader, train_logger, aux_model, projector_s2t=projector_s2t, projector_t2s=projector_t2s)
        
        self.wrt_mode, self.wrt_step = 'train_', 0
        self.log_step = config['trainer'].get('log_per_iter', int(np.sqrt(self.train_loader.batch_size)))
        if config['trainer']['log_per_iter']: self.log_step = int(self.log_step / self.train_loader.batch_size) + 1

        self.num_classes = self.train_loader.dataset.num_classes
        self.svals = [[], []]

        # TRANSORMS FOR VISUALIZATION
        self.restore_transform = transforms.Compose([
            local_transforms.DeNormalize(self.train_loader.MEAN, self.train_loader.STD),
            transforms.ToPILImage()])
        self.viz_transform = transforms.Compose([
            transforms.Resize((400, 400)),
            transforms.ToTensor()])
        
        if self.device ==  torch.device('cpu'): prefetch = False
        if prefetch:
            self.train_loader = DataPrefetcher(train_loader, device=self.device)
            self.val_loader = DataPrefetcher(val_loader, device=self.device)


        # torch.backends.cudnn.benchmark = True
        self.metric_fn = lambda z_s, z_t: F.mse_loss(z_s, z_t)

    def _train_epoch(self, epoch):
        self.logger.info('\n')
            
        self.model.train()
        if self.config['arch']['args']['freeze_bn']:
            if isinstance(self.model, torch.nn.DataParallel): self.model.module.freeze_bn()
            else: self.model.freeze_bn()
        self.wrt_mode = 'train'

        tic = time.time()
        self._reset_metrics()
        tbar = tqdm(self.train_loader, ncols=130)
        for batch_idx, (data, target) in enumerate(tbar):
            self.data_time.update(time.time() - tic)
            #data, target = data.to(self.device), target.to(self.device)

            # target[target == -1] = 255
            # LOSS & OPTIMIZE
            self.optimizer.zero_grad()
            output, z_s1, z_s_lowlevel, z_s2 = self.model(data)

            if self.aux_model is not None:
                with torch.no_grad():
                    match self.config.get("aux_arch"):
                        case "keypointrcnn_rn50":
                            z_t1 = z_t2 = self.aux_model.module.backbone.body(data)['3']
                        case "deeplabv3_rn50":
                            z_t1 = z_t2 = self.aux_model.module.backbone(data)['out']
                        case "segformer" | "segformer-b0" | "segformer-b1":
                            z_t1 = z_t2 = self.aux_model.module.segformer.encoder(data)[0]
                        case _:
                            # Manually running the forward pass of the rn50 without the final avgpool
                            x = self.aux_model.module.conv1(data)
                            x = self.aux_model.module.bn1(x)
                            x = self.aux_model.module.relu(x)
                            x = self.aux_model.module.maxpool(x)

                            x = self.aux_model.module.layer1(x)
                            x = self.aux_model.module.layer2(x)
                            z_t1 = self.aux_model.module.layer3(x)
                            z_t2 = self.aux_model.module.layer4(z_t1) # Ignored to match features

                if self.config.get("no_last_layer_or_interpolate"):
                    z_t = z_t1
                else:
                    z_t = z_t2

            if self.config.get("no_last_layer_or_interpolate"):
                z_s = z_s2
            else:
                z_s = z_s1
            
            proj_loss = torch.tensor(0.0, device=z_s.device)
            t2s_loss = torch.tensor(0.0, device=z_s.device)
            s2t_loss = torch.tensor(0.0, device=z_s.device)

            if self.config['projection_dir'] in ["s2t"]:
                if self.config.get("no_last_layer_or_interpolate"):
                    z_s_proj = self.projector_s2t(z_s)
                else:
                    if not self.config.get("do_pooling"):
                        z_s = F.interpolate(z_s, z_t.shape[-2:])
                    z_s_proj = self.projector_s2t(z_s)
            if self.config['projection_dir'] in ["t2s"]:
                if self.config.get("no_last_layer_or_interpolate"):
                    z_t_proj = self.projector_t2s(z_t)
                else:
                    if not self.config.get("do_pooling"):
                        z_t = F.interpolate(z_t, z_s.shape[-2:])
                    z_t_proj = self.projector_t2s(z_t)

            match self.config['projection_dir']:
                case "t2s":
                    if self.config.get("do_pooling"):
                        b, c, h, w = z_s.shape
                        z_s = z_s.view(b, c, h * w).mean(2)
                        b, c, h, w = z_t_proj.shape
                        z_t_proj = z_t_proj.view(b, c, h * w).mean(2)
                    t2s_loss += self.metric_fn(z_t_proj, z_s)  # Teacher to student
                    proj_loss += t2s_loss
                case "s2t":
                    if self.config.get("do_pooling"):
                        b, c, h, w = z_s_proj.shape
                        z_s_proj = z_s_proj.view(b, c, h * w).mean(2)
                        b, c, h, w = z_t.shape
                        z_t = z_t.view(b, c, h * w).mean(2)
                    s2t_loss += self.metric_fn(z_s_proj, z_t)  # Student to teacher
                    proj_loss += s2t_loss
                case "none":
                    pass
                case _:
                    raise NotImplementedError

            if self.config['arch']['type'][:3] == 'PSP':
                assert output[0].size()[2:] == target.size()[1:]
                assert output[0].size()[1] == self.num_classes 
                task_loss = self.loss(output[0], target)
                task_loss += self.loss(output[1], target) * 0.4
                output = output[0]
            else:
                assert output.size()[2:] == target.size()[1:]
                assert output.size()[1] == self.num_classes 
                task_loss = self.loss(output, target)
            
            loss = proj_loss + task_loss

            if isinstance(self.loss, torch.nn.DataParallel):
                loss = loss.mean()
            loss.backward()
            self.optimizer.step()

            self.lr_scheduler.step()
            self.total_loss.update(loss.item())

            # measure elapsed time
            self.batch_time.update(time.time() - tic)
            tic = time.time()

            # LOGGING & TENSORBOARD
            if batch_idx % self.log_step == 0:
                self.wrt_step = (epoch - 1) * len(self.train_loader) + batch_idx
                self.writer.add_scalar(f'{self.wrt_mode}/loss', loss.item(), self.wrt_step)
                self.writer.add_scalar(f'{self.wrt_mode}/task_loss', task_loss.item(), self.wrt_step)
                if self.projector_s2t is not None:
                    self.writer.add_scalar(f'{self.wrt_mode}/s2t_loss', s2t_loss.item(), self.wrt_step)
                    # svals = torch.linalg.svdvals(self.projector_s2t.module.weight[:, :, 0, 0])
                    # svals_mean = svals.mean()
                    # svals_std = svals.std()

                    # self.writer.add_scalar(f"train/s2t_svals_mean", svals_mean.item(), self.wrt_step)
                    # self.writer.add_scalar(f"train/s2t_svals_std", svals_std.item(), self.wrt_step)
                    # approx_k = torch.count_nonzero(torch.where(svals > (svals_mean + 2 * svals_std), svals, 0.0))
                    # self.writer.add_scalar(f"train/s2t_num_2sd_abv_mean", approx_k, global_step=self.wrt_step)
                    # self.writer.add_histogram(f"train/s2t_svals", svals, global_step=self.wrt_step)

                    # add svd vals for next log (so we can capture the training dynamics of the projection)
                    # self.svals[0] = self.svals[0] + [svals]

                if self.projector_t2s is not None:
                    self.writer.add_scalar(f'{self.wrt_mode}/t2s_loss', t2s_loss.item(), self.wrt_step)
                    # svals = torch.linalg.svdvals(self.projector_t2s.module.weight[:, :, 0, 0])
                    # svals_mean = svals.mean()
                    # svals_std = svals.std()

                    # self.writer.add_scalar(f"train/t2s_svals_mean", svals_mean.item(), self.wrt_step)
                    # self.writer.add_scalar(f"train/t2s_svals_std", svals_std.item(), self.wrt_step)
                    # approx_k = torch.count_nonzero(torch.where(svals > (svals_mean + 2 * svals_std), svals, 0.0))
                    # self.writer.add_scalar(f"train/t2s_num_2sd_abv_mean", approx_k, global_step=self.wrt_step)
                    # self.writer.add_histogram(f"train/t2s_svals", svals, global_step=self.wrt_step)

                    # add svd vals for next log (so we can capture the training dynamics of the projection)
                    # self.svals[1] = self.svals[1] + [svals]

                self.writer.add_scalar(f'{self.wrt_mode}/proj_loss', proj_loss.item(), self.wrt_step)

                for i, opt_group in enumerate(self.optimizer.param_groups):
                    self.writer.add_scalar(f'{self.wrt_mode}/Learning_rate_{i}', opt_group['lr'], self.wrt_step)
                    #self.writer.add_scalar(f'{self.wrt_mode}/Momentum_{k}', opt_group['momentum'], self.wrt_step)


            # FOR EVAL
            seg_metrics = eval_metrics(output, target, self.num_classes)
            self._update_seg_metrics(*seg_metrics)
            pixAcc, mIoU, _ = self._get_seg_metrics().values()
            
            # PRINT INFO
            tbar.set_description(f'TRAIN ({epoch}) ({self.config["projection_dir"]}) | Task Loss: {task_loss:.3f} | Proj Loss: {proj_loss:.3f} | Loss: {self.total_loss.average:.3f} | Acc {pixAcc:.2f} mIoU {mIoU:.2f} | B {self.batch_time.average:.2f} D {self.data_time.average:.2f} |')

        # LOGGING SINGULAR VALUE PLOTS (At end of epoch only)
        if self.projector_s2t is not None:
            svals = torch.linalg.svdvals(self.projector_s2t.module.weight[:, :, 0, 0])
            svals_mean = svals.mean()
            svals_std = svals.std()

            self.writer.add_scalar(f"train/s2t_svals_mean", svals_mean.item(), self.wrt_step)
            self.writer.add_scalar(f"train/s2t_svals_std", svals_std.item(), self.wrt_step)
            approx_k = torch.count_nonzero(torch.where(svals > (svals_mean + 2 * svals_std), svals, 0.0))
            self.writer.add_scalar(f"train/s2t_num_2sd_abv_mean", approx_k, global_step=self.wrt_step)
            self.writer.add_histogram(f"train/s2t_svals", svals, global_step=self.wrt_step)

            # add svd vals for next log (so we can capture the training dynamics of the projection)
            # self.svals[0] = self.svals[0] + [svals]

        if self.projector_t2s is not None:
            svals = torch.linalg.svdvals(self.projector_t2s.module.weight[:, :, 0, 0])
            svals_mean = svals.mean()
            svals_std = svals.std()

            self.writer.add_scalar(f"train/t2s_svals_mean", svals_mean.item(), self.wrt_step)
            self.writer.add_scalar(f"train/t2s_svals_std", svals_std.item(), self.wrt_step)
            approx_k = torch.count_nonzero(torch.where(svals > (svals_mean + 2 * svals_std), svals, 0.0))
            self.writer.add_scalar(f"train/t2s_num_2sd_abv_mean", approx_k, global_step=self.wrt_step)
            self.writer.add_histogram(f"train/t2s_svals", svals, global_step=self.wrt_step)

            # add svd vals for next log (so we can capture the training dynamics of the projection)
            # self.svals[1] = self.svals[1] + [svals]
        
        # if len(self.svals[0]) or len(self.svals[1]) > 1:
            # svd_fig = self.gen_plot(self.svals)
            # self.writer.add_figure('SingularValues', svd_fig, global_step=self.wrt_step)


        # METRICS TO TENSORBOARD
        seg_metrics = self._get_seg_metrics()
        for k, v in list(seg_metrics.items())[:-1]: 
            self.writer.add_scalar(f'{self.wrt_mode}/{k}', v, self.wrt_step)
        

        self.writer.flush()
        # RETURN LOSS & METRICS
        log = {'loss': self.total_loss.average,
                **seg_metrics}

        return log

    def _valid_epoch(self, epoch):
        if self.val_loader is None:
            self.logger.warning('Not data loader was passed for the validation step, No validation is performed !')
            return {}
        self.logger.info('\n###### EVALUATION ######')

        self.model.eval()
        self.wrt_mode = 'val'

        self._reset_metrics()
        tbar = tqdm(self.val_loader, ncols=130)
        with torch.no_grad():
            val_visual = []
            for batch_idx, (data, target) in enumerate(tbar):
                #data, target = data.to(self.device), target.to(self.device)
                # LOSS
                # output = self.model(data)
                # output, z_s, z_s2 = self.model(data)
                output, z_s_nope, z_s2, z_s = self.model(data)
                loss = self.loss(output, target)
                if isinstance(self.loss, torch.nn.DataParallel):
                    loss = loss.mean()
                self.total_loss.update(loss.item())

                seg_metrics = eval_metrics(output, target, self.num_classes)
                self._update_seg_metrics(*seg_metrics)

                # LIST OF IMAGE TO VIZ (15 images)
                if len(val_visual) < 15:
                    target_np = target.data.cpu().numpy()
                    output_np = output.data.max(1)[1].cpu().numpy()
                    val_visual.append([data[0].data.cpu(), target_np[0], output_np[0]])

                # PRINT INFO
                pixAcc, mIoU, _ = self._get_seg_metrics().values()
                tbar.set_description('EVAL ({}) | Loss: {:.3f}, PixelAcc: {:.2f}, Mean IoU: {:.2f} |'.format( epoch,
                                                self.total_loss.average,
                                                pixAcc, mIoU))

            # WRTING & VISUALIZING THE MASKS
            val_img = []
            palette = self.train_loader.dataset.palette
            for d, t, o in val_visual:
                d = self.restore_transform(d)
                t, o = colorize_mask(t, palette), colorize_mask(o, palette)
                d, t, o = d.convert('RGB'), t.convert('RGB'), o.convert('RGB')
                [d, t, o] = [self.viz_transform(x) for x in [d, t, o]]
                val_img.extend([d, t, o])
            val_img = torch.stack(val_img, 0)
            val_img = make_grid(val_img.cpu(), nrow=3, padding=5)
            self.writer.add_image(f'{self.wrt_mode}/inputs_targets_predictions', val_img, self.wrt_step)

            # METRICS TO TENSORBOARD
            self.wrt_step = (epoch) * len(self.val_loader)
            self.writer.add_scalar(f'{self.wrt_mode}/loss', self.total_loss.average, self.wrt_step)
            seg_metrics = self._get_seg_metrics()
            for k, v in list(seg_metrics.items())[:-1]: 
                self.writer.add_scalar(f'{self.wrt_mode}/{k}', v, self.wrt_step)

            log = {
                'val_loss': self.total_loss.average,
                **seg_metrics
            }

        return log

    def gen_plot(self, svals_list_list):
            """Create a pyplot plot and save to buffer."""
            fig, axs = plt.subplots(2, len(svals_list_list), squeeze=False, figsize=[6 * len(svals_list_list), 8])
            [ax.clear() for ax in axs.ravel()]
            labels_idx = ["S->T", "T->S"]
            for svals_idx, svals_list in enumerate(svals_list_list):
                if len(svals_list) == 0:
                    continue
                sns.set_palette("coolwarm", n_colors=len(svals_list))
                axs[0, svals_idx].set_ylim([0, 1.1])
                for i, svals in enumerate(svals_list):
                    svals = svals.cpu().detach().numpy()
                    x = np.linspace(0, len(svals), len(svals))
                    data = pd.DataFrame({"Sorted singular value index": x, "Singular values of Wp": svals})
                    data['Normed Singular values of Wp'] = data['Singular values of Wp'] / data['Singular values of Wp'].max()

                    sns_plot = sns.lineplot(
                        x="Sorted singular value index", 
                        y="Normed Singular values of Wp", 
                        data=data,
                        ax=axs[0, svals_idx]
                    ).set(title=f"Evolvement of singular values of Wp (proj. {labels_idx[svals_idx]})")

                    sns_plot = sns.lineplot(
                        x="Sorted singular value index", 
                        y="Singular values of Wp", 
                        data=data,
                        ax=axs[1, svals_idx]
                    ).set(title=f"Evolvement of singular values of Wp (unnormed) (proj. {labels_idx[svals_idx]})")
                    # ylabels = ['{:,g}'.format(sval) for sval in axs[0, svals_idx].get_yticks()]
                    # axs[0, svals_idx].set_yticklabels(ylabels)

            fig.tight_layout()
            return fig

    def _reset_metrics(self):
        self.batch_time = AverageMeter()
        self.data_time = AverageMeter()
        self.total_loss = AverageMeter()
        self.total_inter, self.total_union = 0, 0
        self.total_correct, self.total_label = 0, 0

    def _update_seg_metrics(self, correct, labeled, inter, union):
        self.total_correct += correct
        self.total_label += labeled
        self.total_inter += inter
        self.total_union += union

    def _get_seg_metrics(self):
        pixAcc = 1.0 * self.total_correct / (np.spacing(1) + self.total_label)
        IoU = 1.0 * self.total_inter / (np.spacing(1) + self.total_union)
        mIoU = IoU.mean()
        return {
            "Pixel_Accuracy": np.round(pixAcc, 3),
            "Mean_IoU": np.round(mIoU, 3),
            "Class_IoU": dict(zip(range(self.num_classes), np.round(IoU, 3)))
        }